/*
 * Copyright (c) 2018-2022 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *                                           Anthony Mallet on Thu Oct  4 2018
 */
#include "acgps.h"

#include <sys/time.h>

#include <ctype.h>
#include <err.h>
#include <math.h>
#include <poll.h>
#include <stdlib.h>
#include <stdio.h>
#include <strings.h>

#include "codels.h"

static void *		tersus_match(struct gps_channel_s *chan);
static int		tersus_configure(struct gps_device_s *dev);
static int		tersus_unconfigure(struct gps_device_s *dev);
static int		tersus_decode(struct gps_device_s *chan,
                                gps_solution_s *solution,
                                sequence_octet *rtk_data);
static int		tersus_ctl(struct gps_device_s *dev,
                                enum gps_ctl_s request, ...);

const struct gps_driver_s tersus_driver = {
  .protocol = "Tersus",

  .match = tersus_match,
  .configure = tersus_configure,
  .unconfigure = tersus_unconfigure,
  .decode = tersus_decode,
  .ctl = tersus_ctl
};

struct tersus_driver_data_s {
  char model[32];		/* device model */
  uint8_t port;			/* connected com port */

  int16_t utcoffset;		/* #leap seconds since 6 Jan 1980 */
  double period;		/* solution output rate */

  double tdelay;		/* estimated BEST_XYZ transmission delay */
};


/* --- Tersus protocol ----------------------------------------------------- */

/* The Tersus protocol is compatible with the Novatel procotol, but with many
 * features missing, and some bugs ...
 */

enum tersus_msgid {
  msgid_BESTXYZ = 241,
  msgid_TIME = 101,
  msgid_VERSION = 37
};

static const char *tersus_portmap[] = {
  [1] = "com1",
  [2] = "com2",
  [3] = "com3"
};

struct __attribute__((packed)) tersus_version {
  uint32_t	ncomp;
  struct {
    uint32_t	comptype;
    char	model[16];
    char	psn[16];
    char	hwversion[16];
    char	swversion[16];
    char	bootversion[16];
    char	compdate[12];
    char	comptime[12];
  } comp[];
};

struct __attribute__((packed)) tersus_bestxyz {
  uint32_t	solstat;
  uint32_t	postype;
  double	px, py, pz;
  float		pxsigma, pysigma, pzsigma;
  uint32_t	vsolstat;
  uint32_t	veltype;
  double	vx, vy, vz;
  float		vxsigma, vysigma, vzsigma;
  char		stnid[4];
  float		vlatency;
  float		diffage, solage;
  uint8_t	svs, solnsvs, soll1svs, solnmultisvs;
  uint8_t	res[1];
  uint8_t	extsolstat;
  uint8_t	galbdssigmask, gpsglosigmask;
};

struct __attribute__((packed)) tersus_time {
  uint32_t	clockstatus;
  double	offset;
  double	offsetstd;
  double	utcoffset;
  uint32_t	utcyear;
  uint8_t	utcmonth, utcday, utchour, utcmin;
  uint32_t	utcms;
  uint32_t	utcstatus;
};

struct __attribute__((packed)) tersus_hdr {
  uint8_t	sync1, sync2, sync3;
  uint8_t	hlen;
  uint16_t	msgid;
  uint8_t	msgtype;
  uint8_t	port;
  uint16_t	msglen;
  uint16_t	seq;
  uint8_t	idle;
  uint8_t	timestatus;
  uint16_t	week;
  uint32_t	ms;
  uint32_t	status;
  uint16_t	res[1];
  uint16_t	version;
};

union __attribute__((packed)) tersus_msg {
  struct __attribute__((packed)) {
    struct tersus_hdr hdr;
    union {
      struct tersus_bestxyz bestxyz;
      struct tersus_time time;
      struct tersus_version version;
    };
  };

  struct __attribute__((packed)) {
    struct tersus_hdr hpayload;
    uint8_t payload[];
  };
};


static int	tersus_ack_msg(struct gps_channel_s *chan,
                        enum tersus_msgid respid,
                        const union tersus_msg **resp, int timeout,
                        const char *fmt, ...);
static size_t	tersus_recv_msg(struct gps_channel_s *chan,
                        const struct timeval *deadline,
                        const union tersus_msg **log, const uint8_t **rtcm,
                        uint16_t *rtcmid);
static uint32_t	tersus_crc32(uint32_t crc, const void *data, size_t len);


/* --- tersus_match ------------------------------------------------------- */

static void *
tersus_match(struct gps_channel_s *chan)
{
  struct tersus_driver_data_s *ddata;
  const union tersus_msg *msg;
  struct timeval deadline;
  int s;

  /* try for at most 1s */
  gettimeofday(&deadline, NULL);
  deadline.tv_sec += 1;

  /* binary command so that we can do robust matching on a binary header */
  s = gpstty_sendmsg(chan, "\r\nlog versionb\r\n");
  if (s) return NULL;

  do {
    if (tersus_recv_msg(chan, &deadline, &msg, NULL, NULL) == 0)
      return NULL;

    /* match versionb message */
    if (msg->hdr.msgid != msgid_VERSION) continue;

    /* check port map */
    if (msg->hdr.port >= sizeof(tersus_portmap)/sizeof(tersus_portmap[0]) ||
        !tersus_portmap[msg->hdr.port]) {
      warnx("unknown tersus port %d", msg->hdr.port);
      return NULL;
    }

    /* match */
    ddata = malloc(sizeof(*ddata));
    if (!ddata) return NULL;

    snprintf(ddata->model, sizeof(ddata->model),
             "%.16s", msg->version.comp[0].model);
    ddata->port = msg->hdr.port;
    ddata->utcoffset = 0.;
    ddata->period = 0.;
    return ddata;
  } while(1);

  return NULL;
}


/* --- tersus_configure --------------------------------------------------- */

static int
tersus_configure(struct gps_device_s *dev)
{
  struct tersus_driver_data_s *ddata = dev->data;
  const union tersus_msg *msg;
  int s;

  /* delete all logs */
  s = tersus_unconfigure(dev);
  if (s) return s;

  /* UTC offset */
  s = tersus_ack_msg(&dev->channel, msgid_TIME, &msg, 500, "log timeb\r\n");
  if (s) return s;
  ddata->utcoffset = floor(msg->time.utcoffset);

  /* WGS84 */
  s = tersus_ack_msg(&dev->channel, 0, NULL, 500, "datum wgs84\r\n");
  if (s) return s;

  s = tersus_ack_msg(&dev->channel, 0, NULL, 500, "undulation user 0.\r\n");
  if (s) return s;


  /* periodic messages */
  {
    /* max messages per second for 50% bandwidth */
    double mps = dev->channel.baud / (10. /*8N1*/) / (2. /*50%*/) /
      (sizeof(struct tersus_bestxyz));

    if (mps > 20) ddata->period = 0.05;
    else if (mps > 10) ddata->period = 0.1;
    else if (mps > 5) ddata->period = 0.2;
    else if (mps > 4) ddata->period = 0.25;
    else if (mps > 2) ddata->period = 0.5;
    else if (mps > 1) ddata->period = 1;
    else ddata->period = 1/mps;

    s = tersus_ack_msg(
      &dev->channel, 0, NULL, 500, "log %s bestxyzb ontime %g\r\n",
      tersus_portmap[ddata->port], ddata->period);
    if (s) return s;
  }

  ddata->tdelay =
    (sizeof(struct tersus_hdr) + sizeof(struct tersus_bestxyz) + 4) *
    10./*8N1*//dev->channel.baud;

  return 0;
}


/* --- tersus_unconfigure -------------------------------------------------- */

static int
tersus_unconfigure(struct gps_device_s *dev)
{
  struct tersus_driver_data_s *ddata = dev->data;
  int s;

  /* delete all logs */
  s = tersus_ack_msg(&dev->channel, 0, NULL, 500,
                      "unlogall %s\r\n", tersus_portmap[ddata->port]);
  if (s) return s;

  return 0;
}


/* --- tersus_decode ------------------------------------------------------ */

static int
tersus_decode(struct gps_device_s *dev, gps_solution_s *solution,
              sequence_octet *rtk_data)
{
  struct tersus_driver_data_s *ddata = dev->data;
  const union tersus_msg *msg;
  const uint8_t *rtcm;
  uint16_t rtcmid;
  struct timeval tv;
  size_t len;

  while(1) {
    gettimeofday(&tv, NULL);
    len = tersus_recv_msg(&dev->channel, NULL, &msg, &rtcm, &rtcmid);
    if (len == 0) break;

    if (rtcm) {
      if (rtk_data->_maximum < rtk_data->_length + len)
        if (genom_sequence_reserve(rtk_data, rtk_data->_length + len))
          return (errno = ENOMEM);
      memcpy(&rtk_data->_buffer[rtk_data->_length], rtcm, len);
      rtk_data->_length += len;
    }

    if (!msg) continue;
    switch(msg->hdr.msgid) {
      case msgid_BESTXYZ: {
        const struct tersus_bestxyz *bestxyz = &msg->bestxyz;
        or_time_ts utc, ts;

        utc = gpsconv_gpsutc(msg->hdr.week, msg->hdr.ms, ddata->utcoffset);
        ts = gpsconv_tvts(tv, - ddata->tdelay);

        solution->ts = ts;
        solution->utc = utc;
        solution->fix = gps_fix_none; /* by default, but updated below */
        solution->sats = bestxyz->solnsvs;

        solution->vel = (gps_enu_s){
          .east = nan(""), .north = nan(""), .up = nan("")
        };
        solution->vel_sigma = (gps_enu_s){
          .east = 0., .north = 0., .up = 0.
        };

        switch(bestxyz->solstat) {
          case 0:
            switch(bestxyz->postype) {
              case  1: solution->fix = gps_fix_fixed; break;
              case 19: solution->fix = gps_fix_interp; break;
              case 16: solution->fix = gps_fix_single; break;
              case 17: solution->fix = gps_fix_diff; break;

              case 32:
              case 33:
              case 34: solution->fix = gps_fix_rtk_float; break;

              case 48:
              case 50: solution->fix = gps_fix_rtk; break;
            }
            break;
        }

        gpsconv_ecefllh(
          &(gps_ecef_s){
            .x = bestxyz->px, .y = bestxyz->py, .z = bestxyz->pz
          }, &solution->llh);
        gpsconv_ecefenu(
          &(gps_ecef_s){
            .x = bestxyz->px, .y = bestxyz->py, .z = bestxyz->pz
          }, &solution->pos);
        gpsconv_ecefenu_stddev(
          &(gps_ecef_s){
            .x = bestxyz->pxsigma, .y = bestxyz->pysigma, .z = bestxyz->pzsigma
          }, &solution->pos_sigma);

        if (bestxyz->vsolstat == 0 && bestxyz->veltype > 0) {
          /* valid velocity */
          gpsconv_decefenu(
            &(gps_ecef_s){
              .x = bestxyz->vx, .y = bestxyz->vy, .z = bestxyz->vz
            }, &solution->vel);
          gpsconv_ecefenu_stddev(
            &(gps_ecef_s){
              .x = bestxyz->vxsigma,
              .y = bestxyz->vysigma,
              .z = bestxyz->vzsigma
            }, &solution->vel_sigma);
        }

        break;
      }
    }
  }

  return 0;
}


/* --- tersus_ctl --------------------------------------------------------- */

static int
tersus_ctl(struct gps_device_s *dev, enum gps_ctl_s request, ...)
{
  struct tersus_driver_data_s *ddata = dev->data;
  va_list args;
  int s = 0;

  va_start(args, request);
  switch(request) {
    case GPSCTL_BAUD: /* set baud rate */ {
      uint32_t baud = va_arg(args, uint32_t);

      s = gpstty_sendmsg(
        &dev->channel, "com %s %d\r\n", tersus_portmap[ddata->port], baud);
      break;
    }

    case GPSCTL_UNFIX: /* unset fixed position */
      s = tersus_ack_msg(&dev->channel, 0, NULL, 500, "fix none\r\n");
      break;

    case GPSCTL_FIX: /* set fixed position */ {
      struct gps_llh_s fix = *va_arg(args, const struct gps_llh_s *);

      s = tersus_ack_msg(&dev->channel, 0, NULL, 500,
                          "fix position %.12f %.12f %.12f\r\n",
                          fix.latitude, fix.longitude, fix.height);
      break;
    }

    case GPSCTL_INRTCM: /* feed RTCM data */ {
      uint16_t type = va_arg(args, unsigned int);
      uint8_t *data = va_arg(args, uint8_t *);
      size_t len = va_arg(args, size_t);

      if (len == 0) break;

      switch (type) {
        case 1: case 3:
        case 1005: case 1006:
        case 1004: case 1012:
        case 1074: case 1084:
        case 1075: case 1085: case 1125:
          /* send to device */
          s = gpstty_send(
            &dev->channel,
            &(struct iovec){ .iov_base = data, .iov_len = len }, 1);
          break;
      }
      break;
    }

    case GPSCTL_OUTRTCM: /* stream RTCM data */ {
      uint16_t rtcmid = va_arg(args, unsigned int);
      double *period = va_arg(args, double *);
      double *offset = va_arg(args, double *);

      if (*period > 0.) {

        if (*period >= 1) *period = round(*period);
        else *period = 0.5;

        if (*offset >= 1) *offset = round(*offset);
        else if (*offset >= 0.5) *offset = 0.5;
        else if (*offset >= 0.25) *offset = 0.25;
        else if (*offset >= 0.2) *offset = 0.2;
        else if (*offset >= 0.1) *offset = 0.1;
        else *offset = 0;

        /* enable log */
        s = tersus_ack_msg(
          &dev->channel, 0, NULL, 500, "log %s rtcm%d ontime %.2f %.2f\r\n",
          tersus_portmap[ddata->port], rtcmid, period, offset);

      } else {

        /* cancel log */
        s = tersus_ack_msg(
          &dev->channel, 0, NULL, 500, "unlog %s rtcm%d\r\n",
          tersus_portmap[ddata->port], rtcmid);
      }
      break;
    }

    case GPSCTL_Q_PARAM: /* query parameters */ {
      gps_param_s *param = va_arg(args, gps_param_s *);

      snprintf(param->model, sizeof(param->model), "%s", ddata->model);
      param->rate = ddata->period > 0. ? 1./ddata->period : 0.;
      break;
    }

    default: s = 1;
  }
  va_end(args);

  return s;
}


/* --- tersus_ack_msg ----------------------------------------------------- */

static int
tersus_ack_msg(struct gps_channel_s *chan, enum tersus_msgid respid,
                const union tersus_msg **resp, int timeout,
                const char *fmt, ...)
{
  struct timeval deadline;
  const char *ack;
  char emsg[256];
  size_t eol, len;
  ssize_t n;
  int s;

  gettimeofday(&deadline, NULL);
  deadline.tv_usec += timeout * 1000;
  if (deadline.tv_usec > 1000000) {
    deadline.tv_sec += deadline.tv_usec / 1000000;
    deadline.tv_usec = deadline.tv_usec % 1000000;
  }

  if (fmt) {
    va_list args;

    va_start(args, fmt);
    s = gpstty_vsendmsg(chan, fmt, args);
    va_end(args);

    if (s) return s;
  }

  /* ignore all messages until an abbreviated ASCII response */
  while(1) {
    ack = gpstty_data(chan, 0, &len);
    if (len < 1) {
      n = gpstty_recv(chan, &deadline);
      if (n < 0) return 1; /* err */
      if (n == 0) { errno = EAGAIN; return 1; /* not enough data */ }
      continue;
    }

    /* match header */
    if (*ack != '<') {
      gpstty_data(chan, 1, NULL); /* skip 1 byte */
      continue;
    }

    /* get full line */
    if (len > 128) len = 128;
    for(eol = 1; eol < len - 1; eol++) {
      if (ack[eol] == '\r' && ack[eol+1] == '\n') break;

      /* unexpected char, skip */
      if (!isprint((unsigned char)ack[eol]) || ack[eol] == '<') break;
    }
    if (eol >= len - 1) {
      /* need more data */
      n = gpstty_recv(chan, &deadline);
      if (n <= 0) return 1; /* err or not enough data */
      continue;
    }
    if (ack[eol] != '\r') {
      gpstty_data(chan, 1, NULL); /* skip 1 byte */
      continue;
    }
    ack = gpstty_data(chan, eol + 2, NULL);

    /* check errors */
    if (strncasecmp(&ack[1], "OK", 2)) {
      if (eol > sizeof(emsg)) eol = sizeof(emsg);
      snprintf(emsg, eol, "%s", &ack[1]);
      warnx("%s", emsg);
      errno = EIO;
      return 1;
    }

    /* ok */
    break;
  }

  /* expecting response message? */
  if (!resp) return 0;

  /* wait until respid */
  while(1) {
    if (tersus_recv_msg(chan, &deadline, resp, NULL, NULL) == 0) {
      errno = EAGAIN;
      return 1;
    }
    if ((*resp)->hdr.msgid == respid) break;
  }

  return 0;
}


/* --- tersus_recv_msg ---------------------------------------------------- */

static size_t
tersus_recv_msg(struct gps_channel_s *chan, const struct timeval *deadline,
                const union tersus_msg **log, const uint8_t **rtcm,
                uint16_t *rtcmid)
{
  static const struct tersus_hdr expect = {
    .sync1 = 0xaa, .sync2 = 0x44, .sync3 = 0x12
  };

  const struct tersus_hdr *hdr;
  const uint8_t *raw;
  size_t len, req, skipped;
  ssize_t n;

  if (log) *log = NULL;
  if (rtcm) *rtcm = NULL;
  skipped = 0;

  do {
    /* check 10 raw bytes to intercept raw RTCM */
    raw = gpstty_data(chan, 0, &len);
    if (len < 10) {
      n = gpstty_recv(chan, deadline);
      if (n <= 0) break; /* err or not enough data */
      continue;
    }

    if (gpsrtk_is_rtcm(raw, rtcmid, &req)) {
      if (skipped) { warnx("skipped %zu unknown bytes", skipped); skipped = 0; }

      for(; len < req; gpstty_data(chan, 0, &len)) {
        n = gpstty_recv(chan, deadline);
        if (n <= 0) break;
      }
      if (len < req) break; /* err or not enough data */

      raw = gpstty_data(chan, req, &len);
      if (!rtcm) continue; /* skip unwanted rtcm data */

      *rtcm = raw;
      return req;
    }

    /* peek one header */
    hdr = gpstty_data(chan, 0, &len);
    if (len < sizeof(*hdr)) {
      n = gpstty_recv(chan, deadline);
      if (n <= 0) break; /* err or not enough data */
      continue;
    }

    /* match binary sync bytes */
    if (hdr->sync1 == expect.sync1 &&
        hdr->sync2 == expect.sync2 &&
        hdr->sync3 == expect.sync3) {
      if (skipped) { warnx("skipped %zu unknown bytes", skipped); skipped = 0; }

      /* get enough data for the whole message */
      req = hdr->hlen + hdr->msglen + 4;
      for(; len < req; gpstty_data(chan, 0, &len)) {
        n = gpstty_recv(chan, deadline);
        if (n <= 0) break;
      }

      const union tersus_msg *msg = gpstty_data(chan, 0, &len);
      if (len < req) break; /* err or not enough data */

      /* check crc */
      if (tersus_crc32(
            tersus_crc32(
              tersus_crc32(0, &msg->hdr, msg->hdr.hlen),
              msg->payload, msg->hdr.msglen),
            msg->payload + msg->hdr.msglen, 4)) {
        warnx("tersus message %d with bad checksum", msg->hdr.msgid);

        gpstty_data(chan, 3, NULL); /* skip 3 sync bytes */
        skipped += 3;
        continue;
      }

      gpstty_data(chan, req, NULL);
      if (!log) continue;  /* skip unwanted message */

      *log = msg;
      return req;
    }

    gpstty_data(chan, 1, NULL); /* skip 1 byte */
    skipped++;
  } while(1);

  return 0;
}


/* --- tersus_crc32 -------------------------------------------------------- */

static uint32_t
tersus_crc32_value(uint8_t x)
{
  static const uint32_t poly = 0xedb88320l;
  uint32_t crc = x;
  int i;

  for(i = 0; i < 8; i++) {
    if (crc & 1)
      crc = (crc >> 1) ^ poly;
    else
      crc >>= 1;
  }

  return crc;
}

static uint32_t
tersus_crc32(uint32_t crc, const void *data, size_t len)
{
  const uint8_t *p = data;

  while(len--)
    crc = (crc >> 8) ^ tersus_crc32_value(((uint8_t)crc) ^ *p++);

  return crc;
}
