/*
 * Copyright (c) 2018-2022 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *                                           Anthony Mallet on Thu Oct 18 2018
 *                                Antonio Enrique Jimenez Cano on Thu Nov 2019
 */
#include "acgps.h"

#include <sys/time.h>

#include <err.h>
#include <math.h>
#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>
#include <time.h>

#include <poll.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include "codels.h"

static void *		ublox_match(struct gps_channel_s *chan);
static int		ublox_configure(struct gps_device_s *dev);
static int		ublox_unconfigure(struct gps_device_s *dev);
static int		ublox_decode(struct gps_device_s *dev,
                                struct gps_solution_s *solution,
                                sequence_octet *rtk_data);
static int		ublox_ctl(struct gps_device_s *dev,
                                enum gps_ctl_s request, ...);


const struct gps_driver_s ublox_driver = {
  .protocol = "u-blox",

  .match = ublox_match,
  .configure = ublox_configure,
  .unconfigure = ublox_unconfigure,
  .decode = ublox_decode,
  .ctl = ublox_ctl
};


/* --- UBX protocol -------------------------------------------------------- */

enum ublox_msgcid {
  mcid_ACK_ACK = 0x0501,
  mcid_ACK_NAK = 0x0500,

  mcid_CFG_DGNSS = 0x0670,
  mcid_CFG_INF = 0x0602,
  mcid_CFG_MSG = 0x0601,
  mcid_CFG_NAV5 = 0x0624,
  mcid_CFG_ODO = 0x061e,
  mcid_CFG_PMS = 0x0686,
  mcid_CFG_PRT = 0x0600,
  mcid_CFG_RATE = 0x0608,
  mcid_CFG_TMODE3 = 0x0671,

  mcid_INF_ERROR = 0x0400,
  mcid_INF_WARN = 0x0401,
  mcid_INF_NOTICE = 0x0402,
  mcid_INF_TEST = 0x0403,
  mcid_INF_DEBUG = 0x0404,

  mcid_MON_MSGPP = 0x0a06,
  mcid_MON_TXBUF = 0x0a08,
  mcid_MON_VER = 0x0a04,

  mcid_NAV_EOE = 0x0161,
  mcid_NAV_HPPPOSLLH = 0x0114,
  mcid_NAV_PVT = 0x0107,
  mcid_NAV_SAT = 0x0135,

  mcid_RXM_RTCM = 0x0232,

  mcid_TIM_VRFY = 0x0d06,

  mcid_NMEA_GGA = 0xf000,
  mcid_NMEA_GLL = 0xf001,
  mcid_NMEA_GSA = 0xf002,
  mcid_NMEA_GSV = 0xf003,
  mcid_NMEA_RMC = 0xf004,
  mcid_NMEA_VTG = 0xf005,

  mcid_RTCM_1005 = 0xf505,
  mcid_RTCM_1074 = 0xf54a,
  mcid_RTCM_1077 = 0xf54d,
  mcid_RTCM_1084 = 0xf554,
  mcid_RTCM_1087 = 0xf557,
  mcid_RTCM_1094 = 0xf55e,
  mcid_RTCM_1097 = 0xf561,
  mcid_RTCM_1124 = 0xf57c,
  mcid_RTCM_1127 = 0xf57f,
  mcid_RTCM_1230 = 0xf5e6,
  mcid_RTCM_4072_0 = 0xf5fe
};

enum ublox_portid {
  prtid_I2C =	0,
  prtid_UART1 =	1,
  prtid_USB =	3,
  prtid_SPI =	4,

  prtid_INVALID =	255
};

struct __attribute__((packed)) ublox_ack {
  uint8_t msgclass, msgid;
};

struct __attribute__((packed)) ublox_cfg_dgnss {
  uint8_t	dgnssMode; /* 2:RTK-float 3:RTK-fixed */
  uint8_t	res[3];
};

struct __attribute__((packed)) ublox_cfg_inf {
  uint8_t proto;
  uint8_t reserved[3];
  uint8_t mask[6];
};

struct __attribute__((packed)) ublox_cfg_msg {
  uint8_t	msgclass, msgid;
  uint8_t	rate[6];
};

struct __attribute__((packed)) ublox_cfg_nav5 {
  uint16_t	mask;
  uint8_t	dynmodel;
  uint8_t	fixmode;
  int32_t	fixedalt;
  uint32_t	fixedaltvar;
  int8_t	minelev;
  uint8_t	drlimit;
  uint16_t	pdop, tdop;
  uint16_t	pacc, tacc;
  uint8_t	staticholdthresh;
  uint8_t	dgnsstimeout;
  uint8_t	cnothreshnumsvs;
  uint8_t	cnothresh;
  uint8_t	res1[2];
  uint16_t	staticholdmaxdist;
  uint8_t	utcstandard;
  uint8_t	res2[5];
};

struct __attribute__((packed)) ublox_cfg_odo {
  uint8_t	version;
  uint8_t	res1[3];
  uint8_t	flags;
  uint8_t	odocfg;
  uint8_t	res2[6];
  uint8_t	cogmaxspeed, cogmaxposacc;
  uint8_t	res3[2];
  uint8_t	vellpgain, coglpgain;
  uint8_t	res4[2];
};

struct __attribute__((packed)) ublox_cfg_pms {
  uint8_t	version;
  uint8_t	powersetupvalue;
  uint16_t	period;
  uint16_t	ontime;
  uint8_t	res[2];
};

struct __attribute__((packed)) ublox_cfg_prt {
  uint8_t	portid;
  uint8_t	res[1];
  uint16_t	txready;
  union {
    struct {
      uint32_t	mode;
      uint32_t	baud;
      uint16_t	inproto;
      uint16_t	outproto;
      uint16_t	flags;
      uint8_t	res[2];
    } uart;
    struct {
      uint8_t	res1[8];
      uint16_t	inproto;
      uint16_t	outproto;
      uint8_t	res2[4];
    } usb;
  };
};

struct __attribute__((packed)) ublox_cfg_rate {
  uint16_t	measrate;
  uint16_t	navrate;
  uint16_t	timeref;
};

struct __attribute__((packed)) ublox_cfg_tmode3 {
  uint8_t	version;
  uint8_t	res1[1];
  uint16_t	flags;
  union {
    struct { int32_t lat, lon, height; };
    struct { int32_t x, y, z; };
  };
  union {
    struct { int8_t lathp, lonhp, heighthp; };
    struct { int8_t xhp, yhp, zhp; };
  };
  uint8_t	res2[1];
  uint32_t	fixedposacc;
  uint32_t	svinmindur;
  uint32_t	svinacclimit;
  uint8_t	res3[8];
};

struct __attribute__((packed)) ublox_inf {
  char		str[1], rest[]; /* variable size string */
};

struct __attribute__((packed)) ublox_mon_msgpp {
  uint16_t msg1[8];
  uint16_t msg2[8];
  uint16_t msg3[8];
  uint16_t msg4[8];
  uint16_t msg5[8];
  uint16_t msg6[8];
  uint32_t skipped[6];
};

struct __attribute__((packed)) ublox_mon_txbuf {
  uint16_t pending[6];
  uint8_t usage[6];
  uint8_t peakusage[6];
  uint8_t tusage;
  uint8_t tpeakusage;
  uint8_t errors;
  uint8_t res[1];
};

struct __attribute__((packed)) ublox_mon_ver {
  char swver[30];
  char hwver[10];
  struct { char data[30]; } ext[];
};

struct __attribute__((packed)) ublox_nav_eoe {
  uint32_t	itow;
};

struct __attribute__((packed)) ublox_nav_hppposllh {
  uint8_t	version;
  uint8_t	res[3];
  uint32_t	itow;
  int32_t	lon, lat, height, hmsl;
  int8_t	lonhp, lathp, heighthp, hmslhp;

  uint32_t	hacc, vacc;
};

struct __attribute__((packed)) ublox_nav_pvt {
  uint32_t	itow;
  uint16_t	year;
  uint8_t	month, day, hour, min, sec;
  uint8_t	valid;
  uint32_t	tacc;
  int32_t	nano;

  uint8_t	fixtype, flags, flags2;
  uint8_t	numsv;

  int32_t	lon, lat, height, hmsl;
  uint32_t	hacc, vacc;

  int32_t	veln, vele, veld;
  int32_t	gspeed;
  int32_t	headmot;
  uint32_t	sacc, headacc;

  uint16_t	pdop;
  uint8_t	res[6];

  int32_t	headveh;
  int16_t	magdec;
  uint16_t	magacc;
};

struct __attribute__((packed)) ublox_nav_sat {
  uint32_t	itow;
  uint8_t	version;
  uint8_t	numsvs;
  uint8_t	res[2];
  struct {
    uint8_t	gnssid;
    uint8_t	svid;
    uint8_t	cno;
    uint8_t	elev;
    uint16_t	azim;
    uint16_t	prres;
    uint32_t	flags;
  } sv[];
};

struct __attribute__((packed)) ublox_rxm_rtcm {
  uint8_t	version;
  uint8_t	flags;
  uint8_t	res[2];
  uint16_t	refstation;
  uint16_t	msgtype;
};


struct __attribute__((packed)) ublox_hdr {
  uint8_t	sync1, sync2;
  uint8_t	msgclass, msgid;
#define ublox_msgcid16(h) ((uint16_t)(((h)->msgclass << 8) | (h)->msgid))
  uint16_t	msglen;
};

union __attribute__((packed)) ublox_msg {
  struct __attribute__((packed)) {
    struct ublox_hdr hdr;
    union {
      struct ublox_ack ack;
      struct ublox_cfg_dgnss cfg_dgnss;
      struct ublox_cfg_inf cfg_inf;
      struct ublox_cfg_msg cfg_msg;
      struct ublox_cfg_nav5 cfg_nav5;
      struct ublox_cfg_odo cfg_odo;
      struct ublox_cfg_pms cfg_pms;
      struct ublox_cfg_prt cfg_prt;
      struct ublox_cfg_rate cfg_rate;
      struct ublox_cfg_tmode3 cfg_tmode3;
      struct ublox_inf inf;
      struct ublox_mon_msgpp mon_msgpp;
      struct ublox_mon_txbuf mon_txbuf;
      struct ublox_mon_ver mon_ver;
      struct ublox_nav_eoe nav_eoe;
      struct ublox_nav_hppposllh nav_hppposllh;
      struct ublox_nav_pvt nav_pvt;
      struct ublox_nav_sat nav_sat;
      struct ublox_rxm_rtcm rxm_rtcm;
    };
  };

  struct __attribute__((packed)) {
    struct ublox_hdr hpayload;
    uint8_t payload[];
  };
};


/* --- local data ---------------------------------------------------------- */

struct ublox_driver_data_s {
  char model[32];		/* device model */
  double protver;		/* protocol version */
  uint8_t port;			/* current communication port */
  double period;		/* configured solution period */

  uint32_t skipped;		/* skipped bytes by device */
  double tdelay;		/* estimated transmission delay for NAV-* */

  /* last solution messages to be reassembled */
  struct ublox_nav_hppposllh hppposllh;
  struct ublox_nav_pvt pvt;
};

static void	ublox_solution(uint32_t itow, const struct timeval *tv,
                        struct ublox_driver_data_s *ddata,
                        struct gps_solution_s *solution);
static void	ublox_satlist(const struct ublox_nav_sat *nav_sat,
                        struct gps_solution_s *solution);
static int	ublox_ack_msg(struct gps_channel_s *chan,
                        enum ublox_msgcid mcid, void *msg, uint16_t len,
                        int timeout);
static size_t	ublox_recv_msg(struct gps_channel_s *chan,
                        const struct timeval *deadline,
                        const union ublox_msg **ubx,
                        const uint8_t **rtcm, uint16_t *rtcmid);
static int	ublox_send_msg(struct gps_channel_s *chan,
                        enum ublox_msgcid mcid, void *msg, uint16_t len);
static uint16_t	ublox_cksum(const struct ublox_hdr *hdr, const void *data);


/* --- ublox_match --------------------------------------------------------- */

static void *
ublox_match(struct gps_channel_s *chan)
{
  struct ublox_driver_data_s *ddata;
  const union ublox_msg *msg;
  struct timeval deadline;
  unsigned int i;
  int s;

  /* try for at most 500ms */
  gettimeofday(&deadline, NULL);
  deadline.tv_usec += 500 * 1000;
  if (deadline.tv_usec >= 1000000) {
    deadline.tv_sec += deadline.tv_usec / 1000000;
    deadline.tv_usec = deadline.tv_usec % 1000000;
  }

  /* send get version message */
  s = ublox_send_msg(chan, mcid_MON_VER, NULL, 0);
  if (s) return NULL;

  /* receive */
  ddata = NULL;
  do {
    if (ublox_recv_msg(chan, &deadline, &msg, NULL, NULL) == 0)
      return NULL;

    /* match MON-VER message */
    if (ublox_msgcid16(&msg->hdr) != mcid_MON_VER) continue;

    /* match */
    ddata = calloc(1, sizeof(*ddata));
    if (!ddata) return NULL;

    warnx("hw: software version %s", msg->mon_ver.swver);
    warnx("hw: hardware version %s", msg->mon_ver.hwver);
    snprintf(ddata->model, sizeof(ddata->model),
             "u-blox %s", msg->mon_ver.hwver);
    for(i = 0; offsetof(struct ublox_mon_ver, ext[i]) < msg->hdr.msglen; i++) {
      warnx("hw: %s", msg->mon_ver.ext[i].data);
      if (!strncmp(msg->mon_ver.ext[i].data, "MOD=", 4))
        snprintf(ddata->model, sizeof(ddata->model),
                 "u-blox %s", msg->mon_ver.ext[i].data + 4);
      if (!strncmp(msg->mon_ver.ext[i].data, "PROTVER=", 8))
        ddata->protver = strtod(msg->mon_ver.ext[i].data + 8, NULL);
    }
    if (ddata->protver < 15) {
      warnx("hw: unsupported protocol version %f", ddata->protver);
      free(ddata);
      return NULL;
    }

    ddata->port = prtid_INVALID; /* invalid: set below */
    ddata->period = 0.;
  } while(!ddata);

  /* determine the current port - this is cumbersome
   * First set the rate of a message to 255, read back the settings on all
   * ports to find the non-zero one ... */
  struct ublox_cfg_msg cfg;

  /* set 255 rate for TIM-VRFY (supported since protocol 15) */
  cfg = (struct ublox_cfg_msg){
    .msgclass = mcid_TIM_VRFY >> 8,
    .msgid = mcid_TIM_VRFY & 0xff,
    .rate = { 255 }
  };
  s = ublox_ack_msg(chan, mcid_CFG_MSG, &cfg, 3, 500);
  if (s) goto failed;

  /* read settings for all ports (sending only 2 bytes) */
  s = ublox_ack_msg(chan, mcid_CFG_MSG, &cfg, 2, 500);
  if (s) goto failed;

  /* find port */
  for (i=0; i < sizeof(cfg.rate); i++)
    if (cfg.rate[i] == 255) { ddata->port = i; break; }

  /* set back to 0 rate for current port (sending only 3 bytes) */
  cfg.rate[0] = 0;
  s = ublox_ack_msg(chan, mcid_CFG_MSG, &cfg, 3, 500);
  if (s) goto failed;

  /* re-read settings for all ports (sending only 2 bytes) */
  s = ublox_ack_msg(chan, mcid_CFG_MSG, &cfg, 2, 500);
  if (s) goto failed;

  /* check port is ok */
  if (ddata->port == prtid_INVALID) goto failed;
  if (cfg.rate[ddata->port] != 0) goto failed;

  return ddata;

failed:
  if (ddata) free(ddata);
  warnx("detected %s protocol at %dbps, but cannot determine port",
        ublox_driver.protocol, chan->baud);
  return NULL;
}


/* --- ublox_configure ----------------------------------------------------- */

static const struct {
  enum ublox_msgcid msgcid; uint8_t rate;
} default_cfg_msg[] = {
  /* disable default NMEA messages */
  { mcid_NMEA_GGA, 0 },
  { mcid_NMEA_GLL, 0 },
  { mcid_NMEA_GSA, 0 },
  { mcid_NMEA_GSV, 0 },
  { mcid_NMEA_RMC, 0 },
  { mcid_NMEA_VTG, 0 },

  /* disable RTCM messages */
  { mcid_RTCM_1005, 0 },
  { mcid_RTCM_1074, 0 },
  { mcid_RTCM_1077, 0 },
  { mcid_RTCM_1084, 0 },
  { mcid_RTCM_1087, 0 },
  { mcid_RTCM_1094, 0 },
  { mcid_RTCM_1097, 0 },
  { mcid_RTCM_1124, 0 },
  { mcid_RTCM_1127, 0 },
  { mcid_RTCM_1230, 0 },
  { mcid_RTCM_4072_0, 0 },

  /* enable UBX messages */
  { mcid_NAV_EOE, 1 },
  { mcid_NAV_HPPPOSLLH, 1 },
  { mcid_NAV_PVT, 1 },
  { mcid_NAV_SAT, 101 },
  { mcid_RXM_RTCM, 1 },
  { mcid_MON_MSGPP, 7 },
  { mcid_MON_TXBUF, 5 }
};

static int
ublox_configure(struct gps_device_s *dev)
{
  struct ublox_driver_data_s *ddata = dev->data;
  unsigned int i;
  size_t msgsz;
  int s;

  /* port configuration - resend the message thrice as it is sometimes
   * not acked if the baud rate is changed on UART */
  for(i = 0; i < 3; i++) {
    s = ublox_ctl(dev, GPSCTL_BAUD, dev->channel.baud);
    if (s) return s;
    s = ublox_ack_msg(&dev->channel, mcid_CFG_PRT, NULL, 0, 500);
    if (!s) break;
  }
  if (s) {
    warnx("u-blox port configuration failed");
    return s;
  }


  /* periodic messages */
  for(i = 0; i < sizeof(default_cfg_msg)/sizeof(default_cfg_msg[0]); i++) {
    struct ublox_cfg_msg rate = {
      .msgclass = default_cfg_msg[i].msgcid >> 8,
      .msgid = default_cfg_msg[i].msgcid & 0xff,
      .rate[0] = default_cfg_msg[i].rate
    };

    if (default_cfg_msg[i].rate) {
      /* enable on current port */
      s = ublox_ack_msg(&dev->channel, mcid_CFG_MSG, &rate, 3, 500);
    } else {
      /* disable on all ports - otherwise quickly gives txbuf errors */
      /* ignore ack errors - message might be unsupported */
      s = ublox_send_msg(&dev->channel, mcid_CFG_MSG, &rate, sizeof(rate));
    }
    if (s) {
      warnx("u-blox message 0x%04x failed", default_cfg_msg[i].msgcid);
      return s;
    }
  }


  /* information messages */
  struct ublox_cfg_inf inf = {0};

  /* NMEA */
  inf.proto = 1;
  s = ublox_ack_msg(&dev->channel, mcid_CFG_INF, &inf, 1, 500);
  if (s) return s;
  inf.mask[ddata->port] = 0;	/* none */
  s = ublox_ack_msg(
    &dev->channel, mcid_CFG_INF, &inf, sizeof(inf), 500);
  if (s) return s;

  /* UBX */
  inf.proto = 0;
  s = ublox_ack_msg(&dev->channel, mcid_CFG_INF, &inf, 1, 500);
  if (s) return s;
  inf.mask[ddata->port] = 7;	/* NOTICE, WARN, ERR */
  s = ublox_ack_msg(
    &dev->channel, mcid_CFG_INF, &inf, sizeof(inf), 500);
  if (s) return s;


  /* power mode (not allowed on USB: always full power) */
  if (ddata->port != prtid_USB) {
    struct ublox_cfg_pms pms = {
      .powersetupvalue = 0	/* full power */
    };
    s = ublox_ack_msg(
      &dev->channel, mcid_CFG_PMS, &pms, sizeof(pms), 500);
    if (s) return s;
  }


  /* navigation settings */
  struct ublox_cfg_nav5 nav5 = {
    .mask = 5,		/* dyn|posFixMode */
    .dynmodel = 8,	/* airbone with <4g acceleration */
    .fixmode = 2	/* 3D only */
  };
  s = ublox_ack_msg(
    &dev->channel, mcid_CFG_NAV5, &nav5, sizeof(nav5), 500);
  if (s) return s;

  struct ublox_cfg_odo odo = {
    .flags = 0		/* no filter */
  };
  s = ublox_ack_msg(
    &dev->channel, mcid_CFG_ODO, &odo, sizeof(odo), 500);
  if (s) return s;


  /* differential mode */
  struct ublox_cfg_dgnss dgnss = {
    .dgnssMode = 3	/* RTK fixed if possible */
  };
  s = ublox_ack_msg(
    &dev->channel, mcid_CFG_DGNSS, &dgnss, sizeof(dgnss), 500);
  if (s) return s;


  /* computation rate */
  struct ublox_cfg_rate rate = {
    .measrate = 50,	/* 20Hz */
    .navrate = 1,	/* on every measurement */
    .timeref = 0	/* UTC time sync */
  };
  s = ublox_ack_msg(
    &dev->channel, mcid_CFG_RATE, &rate, sizeof(rate), 500);
  if (s) return s;

  ddata->period = 0.05;
  switch(ddata->port) {
    case prtid_UART1: /* delay with 8N1 UART */
      msgsz =
        3 * (sizeof(struct ublox_hdr) + 2 /* cksum */) +
        sizeof(struct ublox_nav_pvt) +
        sizeof(struct ublox_nav_hppposllh) +
        sizeof(struct ublox_nav_eoe);
      ddata->tdelay = msgsz * 10./dev->channel.baud;
      break;
    case prtid_USB: /* assume 1ms delay on USB (frame) */
      ddata->tdelay = 1e-3;
      break;
  }

  return 0;
}


/* --- ublox_unconfigure --------------------------------------------------- */

static int
ublox_unconfigure(struct gps_device_s *dev)
{
  unsigned int i;

  /* periodic messages */
  for(i = 0; i < sizeof(default_cfg_msg)/sizeof(default_cfg_msg[0]); i++) {
    struct ublox_cfg_msg rate = {
      .msgclass = default_cfg_msg[i].msgcid >> 8,
      .msgid = default_cfg_msg[i].msgcid & 0xff,
      .rate = { 0 }
    };

    ublox_ack_msg(&dev->channel, mcid_CFG_MSG, &rate, 3, 500);
  }

  return 0;
}


/* --- ublox_decode -------------------------------------------------------- */

static int
ublox_decode(struct gps_device_s *dev, struct gps_solution_s *solution,
             sequence_octet *rtk_data)
{
  (void)rtk_data;
  struct ublox_driver_data_s *ddata = dev->data;
  const union ublox_msg *msg;
  const uint8_t *rtcm;
  uint16_t rtcmid;
  struct timeval tv;
  size_t len;

  while(1) {
    gettimeofday(&tv, NULL);
    len = ublox_recv_msg(&dev->channel, NULL, &msg, &rtcm, &rtcmid);
    if (len == 0) break;

    if (rtcm) {
      if (rtk_data->_maximum < rtk_data->_length + len)
        if (genom_sequence_reserve(rtk_data, rtk_data->_length + len))
          return (errno = ENOMEM);
      memcpy(&rtk_data->_buffer[rtk_data->_length], rtcm, len);
      rtk_data->_length += len;
    }

    if (!msg) continue;
    switch(ublox_msgcid16(&msg->hdr)) {
      case mcid_RXM_RTCM: {
        const struct ublox_rxm_rtcm *rxm = &msg->rxm_rtcm;

        if (rxm->flags & 1)
          warnx("RTCM message %d had CRC failures", rxm->msgtype);
        break;
      }

      case mcid_MON_MSGPP: {
        const struct ublox_mon_msgpp *pp = &msg->mon_msgpp;

        if (ddata->skipped &&
            ddata->skipped != pp->skipped[ddata->port])
          warnx("Hardware skipped %d bytes",
                pp->skipped[ddata->port] - ddata->skipped);

        ddata->skipped = pp->skipped[ddata->port];
        break;
      }

      case mcid_MON_TXBUF: {
        const struct ublox_mon_txbuf *tx = &msg->mon_txbuf;
        int i;

        for (i = 0; i < 6; i++) {
          if (tx->usage[i] > 90 || tx->peakusage[i] > 90)
            warnx("hw transmit buffer %d: %d bytes, %d%% usage, %d%% peak",
                  i, tx->pending[i], tx->usage[i], tx->peakusage[i]);
        }
        break;
      }

      case mcid_NAV_HPPPOSLLH:
        ddata->hppposllh = msg->nav_hppposllh;
        break;

      case mcid_NAV_PVT:
        ddata->pvt = msg->nav_pvt;
        break;

      case mcid_NAV_SAT:
        ublox_satlist(&msg->nav_sat, solution);
        break;

      case mcid_NAV_EOE:
        ublox_solution(msg->nav_eoe.itow, &tv, ddata, solution);
        return 0 /* publish now */;

      default:
        warnx("unexpected message 0x%04x", ublox_msgcid16(&msg->hdr));
    }
  }

  return EAGAIN;
}

static void
ublox_solution(uint32_t itow, const struct timeval *tv,
               struct ublox_driver_data_s *ddata,
               struct gps_solution_s *solution)
{
  const struct ublox_nav_pvt *pvt = &ddata->pvt;
  const struct ublox_nav_hppposllh *hppposllh = &ddata->hppposllh;
  or_time_ts utc, ts;

  if (itow != pvt->itow) /* no NAV-PVT for this epoch */ return;
  if (!(pvt->valid & 0x7)) /* invalid time data */ return;

  utc = gpsconv_utcmktime(
    &(struct tm){
        .tm_sec = pvt->sec,
        .tm_min = pvt->min,
        .tm_hour = pvt->hour,
        .tm_mday = pvt->day,
        .tm_mon = pvt->month - 1,
        .tm_year = pvt->year - 1900
    }, pvt->nano);
  ts = gpsconv_tvts(*tv, - ddata->tdelay);

  /* default solution from NAV-PVT */
  solution->ts = ts;
  solution->utc = utc;
  solution->fix = gps_fix_none; /* by default, but updated below */
  solution->sats = pvt->numsv;

  solution->llh = (gps_llh_s){
    .latitude = pvt->lat * 1e-7,
    .longitude = pvt->lon * 1e-7,
    .height = pvt->height * 1e-3
  };
  solution->pos_sigma = (gps_enu_s){
    .east = pvt->hacc * 1e-3,
    .north = pvt->hacc * 1e-3,
    .up = pvt->vacc * 1e-3
  };

  solution->vel = (gps_enu_s){
    .east = pvt->vele * 1e-3,
    .north = pvt->veln * 1e-3,
    .up = - pvt->veld * 1e-3
  };
  solution->vel_sigma = (gps_enu_s){
    .east = pvt->sacc * 1e-3,
    .north = pvt->sacc * 1e-3,
    .up = pvt->sacc * 1e-3
  };

  /* fix type */
  switch(pvt->fixtype) {
    case 3: /* 3D fix */
    case 4: /* 3D fix + dead reckoning */
      if (!(pvt->flags & 0x1)) break; /* invalid fix */

      if (pvt->flags & 0x80) solution->fix = gps_fix_rtk;
      else if (pvt->flags & 0x40) solution->fix = gps_fix_rtk_float;
      else if (pvt->flags & 0x2) solution->fix = gps_fix_diff;
      else solution->fix = gps_fix_single;

      break;

    case 5: /* fixed position */
      solution->fix = gps_fix_fixed;
      break;
  }

  /* high precision position */
  if (itow == hppposllh->itow) {
    solution->llh = (gps_llh_s){
      .latitude = hppposllh->lat * 1e-7 + hppposllh->lathp * 1e-9,
      .longitude = hppposllh->lon * 1e-7 + hppposllh->lonhp * 1e-9,
      .height = hppposllh->height * 1e-3 + hppposllh->heighthp * 1e-4,
    };

    solution->pos_sigma = (gps_enu_s){
      .east = hppposllh->hacc * 1e-4,
      .north = hppposllh->hacc * 1e-4,
      .up = hppposllh->vacc * 1e-4
    };
  }

  /* ENU position according to configured reference */
  gpsconv_llhenu(&solution->llh, &solution->pos);
}


/* --- ublox_satlist ------------------------------------------------------- */

static void
ublox_satlist(const struct ublox_nav_sat *nav_sat,
              struct gps_solution_s *solution)
{
  static const char *gnssid[] = {
    [0] = "GPS", [1] = "SBAS", [2] = "Galileo", [3] = "BeiDou",
    [4] = "IMES", [5] = "QZSS", [6] = "GLONASS"
  };

  sequence8_gps_satlist_s *list = &solution->satlist;
  sequence64_unsigned_short *id;
  gps_satlist_s *slot;
  uint8_t i;
  size_t n;

  /* reset */
  list->_length = 0;
  solution->totalsats = nav_sat->numsvs;

  /* scan */
  for(i = 0; i < nav_sat->numsvs; i++) {
    if (nav_sat->sv[i].gnssid >= sizeof(gnssid)/sizeof(*gnssid)) continue;

    /* find gnssid slot or create */
    for(n = 0; n < list->_length; n++)
      if (!strcmp(list->_buffer[n].system, gnssid[nav_sat->sv[i].gnssid]))
        break;
    slot = &list->_buffer[n];

    if (n >= list->_length) {
      if (list->_length >= list->_maximum) continue;
      list->_length++;
      snprintf(slot->system, sizeof(slot->system),
               "%s", gnssid[nav_sat->sv[i].gnssid]);
      slot->used._length = slot->unused._length = 0;
    }

    /* store */
    id = nav_sat->sv[i].flags & 0x8 ? &slot->used : &slot->unused;
    if (id->_length >= id->_maximum) continue;
    id->_buffer[id->_length] = nav_sat->sv[i].svid;
    id->_length++;
  }
}


/* --- ublox_ctl ----------------------------------------------------------- */

static int
ublox_ctl(struct gps_device_s *dev, enum gps_ctl_s request, ...)
{
  struct ublox_driver_data_s *ddata = dev->data;
  va_list args;
  int s = 0;

  va_start(args, request);
  switch(request) {
    case GPSCTL_BAUD: /* set baud rate */ {

      uint32_t baud = va_arg(args, uint32_t);

      struct ublox_cfg_prt prt = {
        .portid = ddata->port,
        .txready = 0		/* tx pin disabled */
      };
      switch(ddata->port) {
        case prtid_UART1:
          prt.uart.mode = (3<<6)|(4<<9);		/* 8N1 */
          prt.uart.baud = baud;				/* baud */
          prt.uart.inproto = (1<<0)|(1<<2)|(1<<5);	/* UBX/RTCM2/RTCM3 */
          prt.uart.outproto = (1<<0)|(1<<5);		/* UBX/RTCM3 */
          prt.uart.flags = 0;				/* 1.5s timeout */
          break;

        case prtid_USB:
          prt.usb.inproto = (1<<0)|(1<<2)|(1<<5);	/* UBX/RTCM2/RTCM3 */
          prt.usb.outproto = (1<<0)|(1<<5);		/* UBX/RTCM3 */
          break;

        default: s = 1; break;
      }
      if (s) break;

      s = ublox_send_msg(&dev->channel, mcid_CFG_PRT, &prt, sizeof(prt));
      break;
    }

    case GPSCTL_UNFIX: /* unset fixed position */ {
      struct ublox_cfg_tmode3 tmode3 = {
        .flags = 0
      };

      s = ublox_ack_msg(
        &dev->channel, mcid_CFG_TMODE3, &tmode3, sizeof(tmode3), 500);
      break;
    }

    case GPSCTL_FIX: /* set fixed position */ {
      struct gps_llh_s fix = *va_arg(args, const struct gps_llh_s *);
      struct gps_enu_s sigma = *va_arg(args, const struct gps_enu_s *);

      struct ublox_cfg_tmode3 tmode3 = {
        .flags = (1<<8)|2, /* lla / fixed mode */
      };
      tmode3.lat = fix.latitude * 1e7;
      tmode3.lon = fix.longitude * 1e7;
      tmode3.height = fix.height * 1e2;
      tmode3.lathp = (fix.latitude * 1e7 - tmode3.lat) * 1e2;
      tmode3.lonhp = (fix.longitude * 1e7 - tmode3.lon) * 1e2;
      tmode3.heighthp = (fix.height * 1e2 - tmode3.height) * 1e2;
      if (sigma.east >= sigma.north && sigma.east >= sigma.up)
        tmode3.fixedposacc = 2 * sigma.east * 1e4;
      else if (sigma.north >= sigma.east && sigma.north >= sigma.up)
        tmode3.fixedposacc = 2 * sigma.north * 1e4;
      else if (sigma.up >= sigma.east && sigma.up >= sigma.north)
        tmode3.fixedposacc = 2 * sigma.up * 1e4;

      s = ublox_ack_msg(
        &dev->channel, mcid_CFG_TMODE3, &tmode3, sizeof(tmode3), 500);
      break;
    }

    case GPSCTL_INRTCM: /* feed RTCM data */ {
      uint16_t type = va_arg(args, unsigned int);
      uint8_t *data = va_arg(args, uint8_t *);
      size_t len = va_arg(args, size_t);

      if (len == 0) break;

      switch (type) {
        /* RTCM2 */
        case 1: case 2:
        case 3: case 9:
          break;

          /* RTCM3 */
        case 1005: case 1006: case 1007: /* base station */
        case 1009: case 1010: case 1011: case 1012: /* observations */
        case 1074: case 1075: case 1077: /* GPS MSM 4, 5, 7 */
        case 1084: case 1085: case 1087: /* GLONASS MSM 4, 5, 7 */
        case 1094: case 1095: case 1097: /* Galileo MSM 4, 5, 7 */
        case 1124: case 1125: case 1127: /* Beidou MSM 4, 5, 7 */
        case 1230: /* GLONASS biases */
        case 4072: /* proprietary ublox PVT (moving base) */
          s = gpstty_send(
            &dev->channel,
            &(struct iovec){ .iov_base = data, .iov_len = len }, 1);
          break;

        default: s = 1;
      }
      break;
    }

    case GPSCTL_OUTRTCM: /* stream RTCM data */ {
      uint16_t rtcmid = va_arg(args, unsigned int);
      double *period = va_arg(args, double *);
      double *offset = va_arg(args, double *);

      enum ublox_msgcid mcid;

      *offset = 0.;
      switch(rtcmid) {
        case 1005: mcid = mcid_RTCM_1005; break;
        case 1074: mcid = mcid_RTCM_1074; break;
        case 1077: mcid = mcid_RTCM_1077; break;
        case 1084: mcid = mcid_RTCM_1084; break;
        case 1087: mcid = mcid_RTCM_1087; break;
        case 1094: mcid = mcid_RTCM_1094; break;
        case 1097: mcid = mcid_RTCM_1097; break;
        case 1124: mcid = mcid_RTCM_1124; break;
        case 1127: mcid = mcid_RTCM_1127; break;
        case 1230: mcid = mcid_RTCM_1230; break;
        case 4072: mcid = mcid_RTCM_4072_0; break;

        default: return *period > 0. ? 1 : 0;
      }

      /* make sure period fits in 8 bits */
      if (*period > 255 * ddata->period) {
        warnx("RTCM %d message period %g greater than maximum %g",
              rtcmid, *period, 255 * ddata->period);
        *period = 255 * ddata->period;
      }

      /* set on current port (sending only 3 bytes) */
      struct ublox_cfg_msg rate = {
        .msgclass = mcid >> 8,
        .msgid = mcid & 0xff,
        .rate[0] = ddata->period > 0. ? *period/ddata->period : 0.
      };
      s = ublox_ack_msg(&dev->channel, mcid_CFG_MSG, &rate, 3, 500);
      break;
    }

    case GPSCTL_HZ: /* set computation rate */ {
      double hz = va_arg(args, double);

      if (hz < 0.02 /* 50000ms period */ || hz > 20 /* hw limit */)
        return 1;

      /* set rate */
      struct ublox_cfg_rate rate = {
        .measrate = 1000/hz,
        .navrate = 1,		/* on every measurement */
        .timeref = 0		/* UTC time sync */
      };
      s = ublox_ack_msg(
        &dev->channel, mcid_CFG_RATE, &rate, sizeof(rate), 500);
      if (s) return s;
      ddata->period = rate.measrate / 1000.;

#if 0 /* XXX */
      /* recompute RTCM messages rate */
      for (i = 0; i < ddata->rtcmstats.sent._length; i++) {
        s = ublox_ctl(dev, GPSCTL_OUTRTCM,
                      ddata->rtcmstats.sent._buffer[i].rtcmid,
                      ddata->rtcmstats.sent._buffer[i].period,
                      ddata->rtcmstats.sent._buffer[i].offset);
      }
#endif
      break;
    }

    case GPSCTL_ECUTOFF: /* set min elevation */ {
      double ecutoff = va_arg(args, double);

      struct ublox_cfg_nav5 nav5 = {
        .minelev = ecutoff,
        .mask = 2
      };
      s = ublox_ack_msg(
        &dev->channel, mcid_CFG_NAV5, &nav5, sizeof(nav5), 500);
      if (s) { errno = EINVAL; return s; }
      break;
    }

    case GPSCTL_Q_PARAM: /* query model */ {
      gps_param_s *param = va_arg(args, gps_param_s *);

      struct ublox_cfg_nav5 nav5;
      s = ublox_ack_msg(&dev->channel, mcid_CFG_NAV5, &nav5, 0, 500);
      if (s) { errno = ETIMEDOUT; return s; }
      if (nav5.mask & 2) {
        param->ecutoff = nav5.minelev;
      }

      snprintf(param->model, sizeof(param->model), "%s", ddata->model);
      param->rate = ddata->period > 0. ? 1./ddata->period : 0.;
      break;
    }

    default: errno = ENOENT; s = 1;
  }
  va_end(args);

  return s;
}


/* --- ublox_ack_msg ------------------------------------------------------- */

static int
ublox_ack_msg(struct gps_channel_s *chan,
              enum ublox_msgcid mcid, void *msg, uint16_t len, int timeout)
{
  const union ublox_msg *ack;
  struct timeval deadline;
  int s;

  gettimeofday(&deadline, NULL);
  deadline.tv_usec += timeout * 1000;
  if (deadline.tv_usec > 1000000) {
    deadline.tv_sec += deadline.tv_usec / 1000000;
    deadline.tv_usec = deadline.tv_usec % 1000000;
  }

  if (msg) {
    s = ublox_send_msg(chan, mcid, msg, len);
    if (s) return s;
  }

  /* ignore all messages until corresponding ACK */
  while(1) {
    if (ublox_recv_msg(chan, &deadline, &ack, NULL, NULL) == 0) {
      warnx("timeout for message 0x%04x", mcid); return -1;
    }

    /* if there is a response, save it back to the message */
    if (ublox_msgcid16(&ack->hdr) == mcid)
      memcpy(msg, ack->payload, ack->hdr.msglen);

    if (ublox_msgcid16(&ack->hdr) != mcid_ACK_ACK &&
        ublox_msgcid16(&ack->hdr) != mcid_ACK_NAK) continue;
    if (ublox_msgcid16(&ack->ack) == mcid)
      break;
  }

  return ublox_msgcid16(&ack->hdr) != mcid_ACK_ACK;
}


/* --- ublox_recv_msg ------------------------------------------------------ */

static size_t
ublox_recv_msg(struct gps_channel_s *chan, const struct timeval *deadline,
               const union ublox_msg **ubx,
               const uint8_t **rtcm, uint16_t *rtcmid)
{
  static const struct ublox_hdr expect = {
    .sync1 = 0xb5, .sync2 = 0x62
  };

  const struct ublox_hdr *hdr;
  const void *raw;
  size_t len, req;
  uint16_t crc;
  ssize_t n;

  if (ubx) *ubx = NULL;
  if (rtcm) *rtcm = NULL;

  do {
    /* peek sizeof(hdr) data */
    hdr = raw = gpstty_data(chan, 0, &len);
    if (len < sizeof(*hdr)) {
      n = gpstty_recv(chan, deadline);
      if (n <= 0) break; /* err or not enough data */
      continue;
    }

    /* check raw RTCM3 (needs 5 bytes, less than sizeof(hdr)) */
    if (gpsrtk_is_rtcm(raw, rtcmid, &req) == 3) {
      for(; len < req; gpstty_data(chan, 0, &len)) {
        n = gpstty_recv(chan, deadline);
        if (n <= 0) break;
      }
      if (len < req) break; /* err or not enough data */

      raw = gpstty_data(chan, req, &len);
      if (!rtcm) continue; /* skip unwanted rtcm data */

      *rtcm = raw;
      return req;
    }

    /* match sync bytes */
    if (hdr->sync1 != expect.sync1 || hdr->sync2 != expect.sync2) {
      gpstty_data(chan, 1, NULL); /* skip 1 byte */
      continue;
    }

    /* get enough data for the whole message */
    req = sizeof(*hdr) + hdr->msglen + 2;
    for(; len < req; gpstty_data(chan, 0, &len)) {
      n = gpstty_recv(chan, deadline);
      if (n <= 0) break;
    }

    const union ublox_msg *msg = gpstty_data(chan, 0, &len);
    if (len < req) break; /* err or not enough data */

    /* check crc */
    crc = ublox_cksum(&msg->hdr, msg->payload);
    if (msg->payload[msg->hdr.msglen] != (crc & 0xff) ||
        msg->payload[msg->hdr.msglen+1] != (crc >> 8)) {
      warnx("u-blox message 0x%02x%02x with bad checksum",
            msg->hdr.msgclass, msg->hdr.msgid);

      gpstty_data(chan, 2, NULL); /* skip 2 bytes */
      continue;
    }

    gpstty_data(chan, req, NULL);

    /* skip and display information messages */
    switch(ublox_msgcid16(&msg->hdr)) {
      case mcid_INF_ERROR:
        warnx("hw error: %.*s", msg->hdr.msglen, msg->inf.str);
        continue;
      case mcid_INF_WARN:
        warnx("hw warning: %.*s", msg->hdr.msglen, msg->inf.str);
        continue;
      case mcid_INF_NOTICE:
        warnx("hw: %.*s", msg->hdr.msglen, msg->inf.str);
        continue;
      case mcid_INF_TEST:
        warnx("hw test: %.*s", msg->hdr.msglen, msg->inf.str);
        continue;
      case mcid_INF_DEBUG:
        warnx("hw debug: %.*s", msg->hdr.msglen, msg->inf.str);
        continue;
    }

    if (!ubx) continue;  /* skip unwanted message */

    *ubx = msg;
    return req;

  } while(1);

  return 0;
}


/* --- ublox_send_msg ------------------------------------------------------ */

static int
ublox_send_msg(struct gps_channel_s *chan,
               enum ublox_msgcid mcid, void *msg, uint16_t len)
{
  struct ublox_hdr hdr = {
    .sync1 = 0xb5, .sync2 = 0x62,
    .msgclass = mcid >> 8,
    .msgid = mcid & 0xff,
    .msglen = len
  };
  uint16_t cksum = ublox_cksum(&hdr, msg);

  struct iovec iov[] = {
    { .iov_base = &hdr, .iov_len = sizeof(hdr) },
    { .iov_base = msg, .iov_len = len },
    { .iov_base = &cksum, .iov_len = sizeof(cksum) }
  };

  return gpstty_send(chan, iov, sizeof(iov)/sizeof(*iov));
}


/* --- ublox_cksum --------------------------------------------------------- */

static uint16_t
ublox_cksum(const struct ublox_hdr *hdr, const void *data)
{
  uint16_t len = hdr->msglen;
  uint8_t cksum_a, cksum_b;
  const uint8_t *p;

  cksum_a  = hdr->msgclass;      cksum_b  = cksum_a;
  cksum_a += hdr->msgid;         cksum_b += cksum_a;
  cksum_a += hdr->msglen & 0xff; cksum_b += cksum_a;
  cksum_a += hdr->msglen >> 8;   cksum_b += cksum_a;

  for(p = data; len--; p++) {
    cksum_a += *p;               cksum_b += cksum_a;
  }

  return (cksum_b << 8) | cksum_a;
}
