/*
 * Copyright (c) 2018-2023 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *                                           Anthony Mallet on Thu Oct  4 2018
 */

#include "acgps.h"

#include <err.h>
#include <math.h>

#include "gps_c_types.h"
#include "codels.h"


/* --- Task ntp --------------------------------------------------------- */


/** Codel gpsntp_start of task ntp.
 *
 * Triggered by gps_start.
 * Yields to gps_pause_start, gps_main.
 * Throws gps_e_sys.
 */
genom_event
gpsntp_start(const gps_ntp_context_s *ntp_context,
             const genom_context self)
{
  (void)self;

  return ntp_context ? gps_main : gps_pause_start;
}


/** Codel gpsntp_main of task ntp.
 *
 * Triggered by gps_main.
 * Yields to gps_main, gps_update, gps_cancel.
 */
genom_event
gpsntp_main(gps_ntp_context_s **ntp_context, const genom_context self)
{
  int s;

  if (!*ntp_context) return gps_cancel;

  s = gpsntp_poll(*ntp_context, 500/*ms*/);
  if (s > 0) return gps_update;
  if (s == 0) return gps_main;

  return gps_cancel;
}


/** Codel gpsntp_update of task ntp.
 *
 * Triggered by gps_update.
 * Yields to gps_main.
 */
genom_event
gpsntp_update(gps_ntp_context_s **ntp_context,
              gps_ntp_offset_s *local_utc, const genom_context self)
{
  (void)self;

  if (*ntp_context)
    gpsntp_offset(*ntp_context, &local_utc->offset, &local_utc->jitter);

  return gps_main;
}


/** Codel gpsntp_cancel of task ntp.
 *
 * Triggered by gps_cancel.
 * Yields to gps_pause_start.
 */
genom_event
gpsntp_cancel(gps_ntp_context_s **ntp_context,
              gps_ntp_offset_s *local_utc, const genom_context self)
{
  gpsntp_stop(ntp_context, local_utc, self);
  return gps_pause_start;
}


/** Codel gpsntp_stop of task ntp.
 *
 * Triggered by gps_stop.
 * Yields to gps_ether.
 */
genom_event
gpsntp_stop(gps_ntp_context_s **ntp_context,
            gps_ntp_offset_s *local_utc, const genom_context self)
{
  (void)self;

  if (*ntp_context) gpsntp_close(*ntp_context);
  *ntp_context = NULL;
  *local_utc = (gps_ntp_offset_s){ .offset = nan(""), .jitter = nan("") };
  return gps_ether;
}


/* --- Activity ntp_stop ------------------------------------------------ */

/** Codel gpsntp_stop of activity ntp_stop.
 *
 * Triggered by gps_start.
 * Yields to gps_ether.
 * Throws gps_e_sys.
 */
/* already defined in task ntp */



/* --- Activity ntp_peer ------------------------------------------------ */

/** Codel gpsntp_peer of activity ntp_peer.
 *
 * Triggered by gps_start.
 * Yields to gps_ether.
 * Throws gps_e_sys.
 */
genom_event
gpsntp_peer(const char host[128], gps_ntp_context_s **ntp_context,
            const genom_context self)
{
  if (!*ntp_context) {
    *ntp_context = gpsntp_open();
    if (!*ntp_context) return gps_e_sys_error("ntp_open", self);
  }

  if (gpsntp_addpeer(*ntp_context, host))
    return gps_e_sys_error(host, self);

  return gps_ether;
}


/* --- Activity ntp_info ------------------------------------------------ */

/** Codel gps_ntp_info of activity ntp_info.
 *
 * Triggered by gps_start.
 * Yields to gps_ether.
 * Throws gps_e_sys.
 */
genom_event
gps_ntp_info(gps_ntp_context_s **ntp_context, gps_ntp_offset_s *offset,
             char **stats, const genom_context self)
{
  const char *i;

  if (!*ntp_context) {
    errno = ENODEV;
    return gps_e_sys_error("ntp_info", self);
  }

  gpsntp_offset(*ntp_context, &offset->offset, &offset->jitter);

  i = gpsntp_stat(*ntp_context);
  if (!i) return gps_e_sys_error("ntp_info", self);

  *stats = strdup(i);
  return gps_ether;
}
