/*
 * Copyright (c) 2022 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *                                           Anthony Mallet on Sat Oct  8 2022
 */
#include <sys/socket.h>

#include <arpa/inet.h>
#include <netinet/in.h>

#include <assert.h>
#include <endian.h>
#include <err.h>
#include <errno.h>
#include <fcntl.h>
#include <math.h>
#include <poll.h>
#include <netdb.h>
#include <stddef.h>
#include <time.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "codels.h"

/* helpers */

#define countof(x)	(sizeof(x)/sizeof(*(x)))

#define forarray(x, a)                                                  \
  for((x) = (a); (x) < (a) + countof(a); (x)++)
#define forptr(x, a, n)                                                 \
  for((x) = (a); (x) < (a) + (n); (x)++)
#define forlist(x, l, next)                                             \
  for((x) = (l); x; (x) = (x)->next)

#define swap(a, b)                                                      \
  do {                                                                  \
    unsigned char tmp[sizeof(a) * (sizeof(a) == sizeof(b) ? 1:-1)];     \
    memcpy(tmp, &(a), sizeof(a));                                       \
    memcpy(&(a), &(b), sizeof(a));                                      \
    memcpy(&(b), tmp, sizeof(a));                                       \
  } while(0)
#define ssort(a, test, op, n)                                           \
  do {                                                                  \
    int i, j;                                                           \
    for (i = 1; i < (n); i++)                                           \
      for (j = i; j > 0 && ((a)[j] test) op ((a)[j-1] test); j--)       \
        swap((a)[j], (a)[j-1]);                                         \
  } while(0)

#define square(x)	((x)*(x))


/* timing */

#define cmpts(t, op, u)                                                 \
  ((t).tv_sec == (u).tv_sec ?                                           \
   (t).tv_nsec op (u).tv_nsec : (t).tv_sec op (u).tv_sec)

static inline struct timespec
addts(struct timespec t, struct timespec u)
{
  t.tv_sec += u.tv_sec;
  t.tv_nsec += u.tv_nsec;
  if (t.tv_nsec >= 1000000000) {
    t.tv_sec++;
    t.tv_nsec -= 1000000000;
  }
  return t;
}

static inline struct timespec
getsystime(void)
{
  struct timespec ts;
  clock_gettime(CLOCK_REALTIME, &ts);
  return ts;
}


/* fixed point - RFC 5905 section 6. */

typedef uint32_t fp32;
typedef uint64_t fp64;
typedef int64_t sfp64;

static inline double fp32tod(fp32 u) { return ldexp(u, -16); }
static inline double sfp64tod(sfp64 u) { return ldexp(u, -32); }

static inline fp64
fp64ton(fp64 fp)
{
  return (union { uint32_t b[2]; fp64 f; }){
    .b = { htonl(fp >> 32), htonl((uint32_t)fp) } }.f;
}

static inline fp64
ntofp64(fp64 fp)
{
  return (union { uint32_t b[2]; fp64 f; }){
    .b = { ntohl(fp >> 32), ntohl((uint32_t)fp) } }.f;
}

static const uint32_t ntp_jan_1970 = 2208988800;

static inline fp64
tstofp64(struct timespec ts)
{
  return (((uint64_t)ts.tv_sec + ntp_jan_1970) << 32)
    + ((uint64_t)ts.tv_nsec << 32)/1000000000;
}

static inline struct timespec
fp64tots(fp64 fp) {
  return (struct timespec){
    .tv_sec = (fp >> 32) - ntp_jan_1970,
    .tv_nsec = ((fp & 0xffffffff) * 1000000000) >> 32
  };
}


/* ntp packet header - RFC 5905 section 7.3. */

struct __attribute__((packed)) pkt {
  uint8_t li_vn_mode;	/* leap indicator, version number, mode */
  uint8_t stratum;	/* stratum */
  int8_t poll;		/* max poll interval */
  int8_t precision;	/* clock precision */
  fp32 rootdelay;	/* roundtrip delay */
  fp32 rootdisp;	/* dispersion */
  char refid[4];	/* reference id */
  fp64 refts;		/* last update time */
  fp64 org;		/* originate time stamp */
  fp64 rec;		/* receive time stamp */
  fp64 xmt;		/* transmit time stamp */

  uint32_t keyid;	/* optional */
  uint8_t digest[128];	/* optional */
};

static inline uint8_t
livnmodetou(uint8_t li, uint8_t vn, uint8_t mode)
{
  return ((li & 3) << 6) | ((vn & 7) << 3) | (mode & 7);
}

static inline uint8_t ntp_utoli(uint8_t u) { return u >> 6; }
static inline uint8_t ntp_utovn(uint8_t u) { return (u >> 3) & 7; }
static inline uint8_t ntp_utomode(uint8_t u) { return u & 7; }


/* remote NTP peer */

#define NTP_VERSION	3	/* min supported version */
#define NTP_MAXDISP	16.	/* max dispersion (s) */
#define NTP_MAXSTRAT	16
#define	NTP_MINCLOCK	3	/* min number of survivors */
#define NTP_BURST	8
#define NTP_BPOLL	1	/* 2s */
#define NTP_MINPOLL	5	/* 32s */
#define NTP_MAXPOLL	10	/* 17min */

#define NTP_SHIFT	8	/* filter stages */

struct peer {
  struct peer *next, *prev;

  char hname[NI_MAXHOST];	/* canonical name (info messages only) */
  union {			/* socket-address for sendto(2) */
    struct sockaddr_storage addr_storage;
    struct sockaddr addr[1];
  };
  socklen_t addrlen;

  fp64 xmt, rcv;		/* last transmit/receive timestamp */
  struct filter {
    double offset;
    double delay;
    double disp;
    fp64 ts;
  } filter[NTP_SHIFT];		/* packet filter */
  int shift;

  fp64 ts;			/* update time */
  uint8_t stratum;
  double precision;
  double rootdist;
  double delay, rootdelay;
  double disp, rootdisp;

  struct endpt {                /* correctness interval for clock_select() */
    struct peer * const peer;   /* self */
    const int type;             /* +1 upper/-1 lower bound */
    double edge;                /* lower/upper bound */
  } mindist, maxdist;

  double offset;
  double jitter, seljit;

  uint8_t reach, keep;
  int burst;
  int poll;
  struct timespec nextpoll;
};

static int		transmit(gps_ntp_context_s *ctx, struct peer *peer);
static int		receive(gps_ntp_context_s *ctx, int fd);

static void		peer_process(gps_ntp_context_s *ctx,
                                struct peer *peer, fp64 rcv, struct pkt *pkt);
static void		clock_filter(gps_ntp_context_s *ctx,
                                struct peer *peer, double offset, double delay,
                                double disp);

static struct peer *	add_peer(gps_ntp_context_s *ctx,
                                const struct sockaddr *addr, socklen_t len);
static void		del_peer(gps_ntp_context_s *ctx, struct peer *peer,
                                const char *msg);
static struct peer *	find_peer(gps_ntp_context_s *ctx,
                                const struct sockaddr *addr);


/* global context */

#define CTX_MAXPEERS		8	/* maximum number of active peers */
#define CTX_DNSPOLL		300	/* DNS poll interval (s) */
#define CTX_DISP		15e-6	/* 15ppm */

struct gps_ntp_context_s {
  struct pollfd sock[2];	/* AF_INET{,6} sockets */

  char *servers[CTX_MAXPEERS];	/* NTP servers */
  int nservers;
  struct timespec nextdns;	/* next addr resolution request */

  struct peer *peers;		/* active peers */
  int npeers;

  struct peer *best;		/* selected peer for synchronization */
  double offset;
  double jitter;

  fp64 now;			/* current time */
  double precision;		/* rho, local clock precision (s) */
  double disp;			/* phi, local clock dispersion rate (s/s) */

  char *stat;			/* statistics buffer */
  size_t statlen;
};

static inline struct pollfd *
ctxaftosock(gps_ntp_context_s *ctx, sa_family_t af)
{
  return
    af == AF_INET6 ? &ctx->sock[0]:
    af == AF_INET ? &ctx->sock[1]:
    NULL;
}

static void		clock_select(gps_ntp_context_s *ctx);
static void		clock_combine(gps_ntp_context_s *ctx,
                                struct peer **peers, int npeers);
static void		poll_update(struct peer *peer);
static int		lookup_servers(gps_ntp_context_s *ctx);


/* --- gpsntp_open --------------------------------------------------------- */

gps_ntp_context_s *
gpsntp_open()
{
  gps_ntp_context_s *ctx;
  struct addrinfo *res, *res0;
  struct pollfd *pfd;
  int fd, s;

  /* empty context */
  ctx = malloc(sizeof(*ctx));
  if (!ctx) return NULL;
  *ctx = (gps_ntp_context_s){
    .precision = 0.5e-6,		/* minimum */
    .disp = CTX_DISP
  };

  forarray(pfd, ctx->sock) {
    pfd->fd = -1;
    pfd->events = POLLIN;
  }
  ctx->nextdns = getsystime();
  ctx->now = tstofp64(ctx->nextdns);

  /* lookup localhost sockets */
  s = getaddrinfo(
    NULL, "0", &(struct addrinfo){
      .ai_flags = AI_PASSIVE | AI_NUMERICSERV,
        .ai_family = AF_UNSPEC,
        .ai_socktype = SOCK_DGRAM,
        .ai_protocol = IPPROTO_UDP }, &res0);
  if (s) {
    warnx("localhost: %s", gai_strerror(s));
    errno = EINVAL;
    goto err;
  }

  forlist(res, res0, ai_next) {
    if (!ctxaftosock(ctx, res->ai_family)) continue;

    fd = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
    if (fd < 0) continue;

    s = fcntl(fd, F_GETFL, 0);
    if (s == -1) goto errmsg;
    if (fcntl(fd, F_SETFL, s | O_NONBLOCK)) goto errmsg;
    if (res->ai_family == AF_INET6) {
      s = setsockopt(fd, IPPROTO_IPV6, IPV6_V6ONLY, &(int){1}, sizeof(int));
      if (s < 0) goto errmsg;
    }

    /* use timestamps if available */
#ifdef SO_TIMESTAMP
    s = setsockopt(fd, SOL_SOCKET, SO_TIMESTAMP, &(int){1}, sizeof(int));
    if (s < 0) goto errmsg;
#endif

    ctxaftosock(ctx, res->ai_family)->fd = fd;
  }
  freeaddrinfo(res0);

  return ctx;
errmsg:
  warn("ntp_open");
err:
  gpsntp_close(ctx);
  return NULL;
}


/* --- gpsntp_close -------------------------------------------------------- */

void
gpsntp_close(gps_ntp_context_s *ctx)
{
  struct pollfd *s;
  struct peer *p;
  char **h;

  if (!ctx) return;

  if (ctx->stat) free(ctx->stat);
  forarray(s, ctx->sock) if (s->fd >= 0) close(s->fd);
  forarray(h, ctx->servers) if (*h) free(*h);

  while ((p = ctx->peers)) {
    ctx->peers = ctx->peers->next;
    free(p);
  }

  free(ctx);
}


/* --- gpsntp_addpeer ------------------------------------------------------ */

int
gpsntp_addpeer(gps_ntp_context_s *ctx, const char *host)
{
  ctx->servers[ctx->nservers] = strdup(host);
  if (!ctx->servers[ctx->nservers]) return -1;

  ctx->nservers++;
  return lookup_servers(ctx);
}


/* --- gpsntp_offset ------------------------------------------------------- */

void
gpsntp_offset(gps_ntp_context_s *ctx, double *offset, double *jitter)
{
  if (*offset) *offset = ctx->best ? ctx->offset : nan("");
  if (*jitter) *jitter = ctx->best ? ctx->jitter : nan("");
}


/* --- gpsntp_poll --------------------------------------------------------- */

int
gpsntp_poll(gps_ntp_context_s *ctx, int timeout)
{
  struct timespec ts, next;
  struct pollfd *sock;
  struct peer *p;
  int s;

  /* compute next deadline */
  ts = getsystime();
  if (timeout) {
    if (timeout < 0)
      next = ctx->nextdns;
    else {
      next = addts(ts, (struct timespec){
          .tv_sec = timeout/1000, .tv_nsec = (timeout%1000) * 1000000 });
      if (cmpts(next, >, ctx->nextdns)) next = ctx->nextdns;
    }
    forlist(p, ctx->peers, next) {
      if (cmpts(p->nextpoll, ==, (struct timespec){0})) continue;
      if (cmpts(next, >, p->nextpoll)) next = p->nextpoll;
    }
    if (cmpts(next, <=, ts))
      timeout = 0;
    else
      timeout = (next.tv_sec - ts.tv_sec)*1000
                + (next.tv_nsec - ts.tv_nsec)/1000000;
  }

  /* resolve servers */
  if (cmpts(ctx->nextdns, <=, ts)) {
    lookup_servers(ctx);
  }

  /* transmit to peers */
  forlist(p, ctx->peers, next) {
    if (cmpts(p->nextpoll, ==, (struct timespec){0})) continue;

    if (cmpts(p->nextpoll, <=, ts)) transmit(ctx, p);
  }

  /* poll */
  s = poll(ctx->sock, countof(ctx->sock), timeout);
  ts = getsystime();
  ctx->now = tstofp64(ts);
  if (s < 0) return s;
  if (s > 0) {
    forarray(sock, ctx->sock)
      if (sock->revents & POLLIN)
        while (receive(ctx, sock->fd) == 0) /* empty body */;
    clock_select(ctx);
    return 1;
  }

  return 0;
}


/* --- gpsntp_stat --------------------------------------------------------- */

char *
gpsntp_stat(gps_ntp_context_s *ctx)
{
  const struct peer *p;
  const char *end = ctx->stat + ctx->statlen;
  char *b = ctx->stat;
  int s;

increase:
  /* increase buffer size if needed */
  if (b >= end) {
    b = realloc(ctx->stat, ctx->statlen + 1024);
    if (!b) return NULL;
    ctx->stat = b;
    ctx->statlen += 1024;
    end = ctx->stat + ctx->statlen;
  }

#define xprint(...)                                                     \
  do {                                                                  \
    if (b >= end) goto increase;                                        \
    s = snprintf(b, end-b, __VA_ARGS__);                                \
    if (s < 0) return NULL;                                             \
    b += s;                                                             \
  } while(0)

  xprint("\n");
  xprint(" %-23s%-10s%10s%10s%10s%15s\n",
         "peer", "reach/keep", "dist", "delay", "jitter", "offset (ms)");
  forlist(p, ctx->peers, next)
    xprint("%c%-24.23s%4.3o/%-4.3o%10.3f%10.3f%10.3f%+10.3f\n",
           (p == ctx->best) ? '*' : (p->keep & 1) ? '+' : ' ',
           p->hname, p->reach, p->keep, p->rootdist * 1000.,
           p->delay * 1000., p->jitter * 1000., p->offset * 1000.);

#undef xprint
  if (b >= end) goto increase;

  return ctx->stat;
}


/* --- transmit ------------------------------------------------------------ */

static int
transmit(gps_ntp_context_s *ctx, struct peer *peer)
{
  ssize_t l;
  fp64 xmt;

  struct pkt pkt = {
    .li_vn_mode = livnmodetou(0/* ok */, NTP_VERSION, 3/* client */),
    .poll = peer->poll
  };

  xmt = tstofp64(getsystime());
  pkt.xmt = fp64ton(xmt);
  l = sendto(ctxaftosock(ctx, peer->addr->sa_family)->fd,
             &pkt, offsetof(struct pkt, keyid), 0,
             peer->addr, peer->addrlen);

  peer->xmt = xmt;
  peer->reach <<= 1;
  if (peer->burst) peer->burst--;
  else {
    if (!peer->reach)
      del_peer(ctx, peer, "timeout");
    else if (!(peer->reach & 2))
      clock_filter(ctx, peer, 0., NTP_MAXDISP, NTP_MAXDISP);
  }

  poll_update(peer);

  return (l < 0) ? errno : 0;
}


/* --- receive ------------------------------------------------------------- */

static int
receive(gps_ntp_context_s *ctx, int fd)
{
  struct peer *peer;
  struct cmsghdr *h;
  struct pkt pkt;
  ssize_t l;
  fp64 rcv;

  union {
    struct sockaddr_storage storage;
    struct sockaddr addr[1];
  } saddr;

  union {
    struct cmsghdr h;
    char buf[128];
  } cmsg;

  struct iovec iov = {
    .iov_base = &pkt, .iov_len = sizeof(pkt)
  };
  struct msghdr msg = {
    .msg_name = saddr.addr, .msg_namelen = sizeof(saddr),
    .msg_iov = &iov, .msg_iovlen = 1,
    .msg_control = &cmsg, .msg_controllen = sizeof(cmsg)
  };

  l = recvmsg(fd, &msg, 0);
  if (l < 0) return errno;

  /* get peer */
  peer = find_peer(ctx, saddr.addr);
  if (!peer) {
    char name[NI_MAXHOST];
    getnameinfo(saddr.addr, msg.msg_namelen, name, sizeof(name), NULL, 0, 0);
    warnx("unexpected ntp packet from %s", name);
    return 0;
  }

  /* sanity check packet */
  if (msg.msg_flags & MSG_TRUNC) return 0;
  if ((size_t)l < offsetof(struct pkt, keyid)) return 0;
  if (ntp_utovn(pkt.li_vn_mode) != NTP_VERSION) return 0;
  if (ntp_utomode(pkt.li_vn_mode) != 4) return 0;
  if (ntp_utoli(pkt.li_vn_mode) == 3) return 0;
  if (pkt.stratum >= NTP_MAXSTRAT) return 0;

  /* check KoD */
  if (pkt.stratum == 0) {
    if (!memcmp(pkt.refid, "RATE", 4)) {
      peer->burst = 0;
      peer->poll++;
      warnx("rate limiting ntp on %s", peer->hname);
    } else if (!memcmp(pkt.refid, "DENY", 4) || !memcmp(pkt.refid, "RSTR", 4))
      del_peer(ctx, peer, "kicked out");

    return 0;
  }

  /* decode timestamps */
  pkt.rootdelay = ntohl(pkt.rootdelay);
  pkt.rootdisp = ntohl(pkt.rootdisp);
  pkt.org = ntofp64(pkt.org);
  pkt.rec = ntofp64(pkt.rec);
  pkt.xmt = ntofp64(pkt.xmt);
  if (pkt.org != peer->xmt) return 0;

  rcv = ctx->now;
#ifdef SCM_TIMESTAMP
  for (h = CMSG_FIRSTHDR(&msg); h; h = CMSG_NXTHDR(&msg, h)) {
    if (h->cmsg_len != CMSG_LEN(sizeof(struct timeval))) continue;
    if (h->cmsg_level != SOL_SOCKET) continue;
    if (h->cmsg_type != SCM_TIMESTAMP) continue;

    rcv = tstofp64((struct timespec){
        .tv_sec = ((struct timeval *)CMSG_DATA(h))->tv_sec,
        .tv_nsec = ((struct timeval *)CMSG_DATA(h))->tv_usec * 1000 });
    break;
  }
#endif


  /* process packet */
  peer_process(ctx, peer, rcv, &pkt);
  return 0;
}


/* --- peer_process -------------------------------------------------------- */

static void
peer_process(gps_ntp_context_s *ctx,
             struct peer *peer, fp64 rcv, struct pkt *pkt)
{
  double offset, delay, disp;

  peer->reach |= 1;
  peer->rcv = rcv;
  peer->stratum = pkt->stratum;
  peer->precision = ldexp(1., pkt->precision);
  peer->rootdelay = fp32tod(pkt->rootdelay);
  peer->rootdisp = fp32tod(pkt->rootdisp);
  if (!peer->burst && pkt->poll > peer->poll) peer->poll = pkt->poll;

  offset = (sfp64tod(pkt->rec - pkt->org) + sfp64tod(pkt->xmt - peer->rcv))/2.;
  delay = fabs(sfp64tod(pkt->rec - pkt->org) - sfp64tod(pkt->xmt - peer->rcv));
  disp = ctx->precision + peer->precision + ctx->disp * delay;

  clock_filter(ctx, peer, offset, delay, disp);
}


/* --- clock_filter -------------------------------------------------------- */

static void
clock_filter(gps_ntp_context_s *ctx, struct peer *peer,
             double offset, double delay, double disp)
{
  int idx[NTP_SHIFT];
  int i;

  /* shift sample in */
  peer->filter[peer->shift] = (struct filter){
    .offset = offset, .delay = delay, .disp = disp, .ts = ctx->now
  };
  peer->shift = (peer->shift + 1) % NTP_SHIFT;

  /* sort by delay */
  for (i = 0; i < NTP_SHIFT; i++) idx[i] = i;
  ssort(idx, [peer->filter].delay, <, NTP_SHIFT);

  /* use only a sample if it is more recent than the last used sample */
  if ((sfp64)(peer->filter[idx[0]].ts - peer->ts) <= 0)
    return;

  /* update peer variables */
  peer->ts = peer->filter[idx[0]].ts;

  peer->offset = peer->filter[idx[0]].offset;
  peer->delay = peer->filter[idx[0]].delay;

  peer->disp = peer->filter[idx[0]].disp / 2.;
  peer->jitter = 0.;
  for (i = 1; i < NTP_SHIFT; i++) {
    peer->disp += ldexp(peer->filter[idx[i]].disp, -(i+1));
    peer->jitter +=
      square(peer->filter[idx[i]].offset - peer->filter[idx[0]].offset);
  }
  peer->jitter = sqrt(peer->jitter / (NTP_SHIFT-1));
  if (peer->jitter < ctx->precision) peer->jitter = ctx->precision;
}


/* --- clock_select -------------------------------------------------------- */

static void
clock_select(gps_ntp_context_s *ctx)
{
  struct endpt *endpt[2 * CTX_MAXPEERS];
  struct peer *keep[CTX_MAXPEERS];
  struct peer *peer, **p, **q, **w;
  double lowpt, hipt, min, max, err;
  int nkeep;
  int bad, n, i;

  /* push active peers */
  nkeep = 0; forlist(peer, ctx->peers, next) {
    peer->keep <<= 1;
    if (cmpts(peer->nextpoll, ==, (struct timespec){0})) continue;
    assert(nkeep < CTX_MAXPEERS);
    keep[nkeep++] = peer;
  }
  if (!nkeep) return;


  /* build correctness interval list - Section 11.2.1 */
  n = 0; forptr(p, keep, nkeep) {
    (*p)->rootdist =
      ((*p)->delay + (*p)->rootdelay)/2
      + (*p)->disp + (*p)->rootdisp + ctx->disp * sfp64tod(ctx->now - (*p)->ts)
      + (*p)->jitter;
    if ((*p)->rootdist < ctx->precision) (*p)->rootdist = ctx->precision;

    (*p)->mindist.edge = (*p)->offset - (*p)->rootdist;
    (*p)->maxdist.edge = (*p)->offset + (*p)->rootdist;

    endpt[n] = &(*p)->mindist;
    endpt[2*nkeep - n - 1] = &(*p)->maxdist;
    n++;
  }

  ssort(endpt, ->edge, <, 2*nkeep);

  lowpt = 2e9, hipt = -2e9;
  for (bad = 0; 2*bad < nkeep; bad++) {
    for (i = n = 0; i < 2*nkeep; i++) {
      n -= endpt[i]->type;
      if (n >= nkeep - bad) { lowpt = endpt[i]->edge; break; }
    }
    for (i = 2*nkeep-1, n = 0; i >= 0; i--) {
      n += endpt[i]->type;
      if (n >= nkeep - bad) { hipt = endpt[i]->edge; break; }
    }
    if (lowpt < hipt) break;
  }
  if (lowpt >= hipt) return;


  /* remove falsetickers */
  n = 0; forptr(p, keep, nkeep) {
    if ((*p)->mindist.edge > hipt || (*p)->maxdist.edge < lowpt)
      continue;
    keep[n++] = *p;
  }
  nkeep = n;


  /* cluster */
  while (1) {
    w = keep, min = 2e9, max = -2e9;
    forptr(p, keep, nkeep) {
      if ((*p)->jitter < min) min = (*p)->jitter;

      (*p)->seljit = 0.;
      if (nkeep > 1) {
        forptr(q, keep, nkeep)
          (*p)->seljit += square((*q)->offset - (*p)->offset);
        (*p)->seljit = sqrt((*p)->seljit / (nkeep - 1));
      }
      err = (*p)->seljit * (*p)->rootdist;
      if (err > max) { max = err; w = p; }
    }
    max = (*w)->seljit;
    if (max < ctx->precision) max = ctx->precision;
    if (nkeep <= NTP_MINCLOCK || max <= min) break;

    nkeep--;
    forptr(p, w, &keep[nkeep] - w) p[0] = p[1];
  }

  /* combine */
  min = 2e9;
  forptr(p, keep, nkeep) {
    (*p)->keep |= 1;
    err = (*p)->seljit * (*p)->rootdist + (*p)->stratum * 0.001;
    if (err < min) { min = err; peer = *p; }
  }
  ctx->best = peer;

  clock_combine(ctx, keep, nkeep);
}


/* --- clock_combine ------------------------------------------------------- */

static void
clock_combine(gps_ntp_context_s *ctx, struct peer **peers, int npeers)
{
  struct peer **p;
  double x, y, z, w;

  y = z = w = 0.;
  forptr(p, peers, npeers) {
    x = 1. / (*p)->rootdist;

    y += x;
    z += x * (*p)->offset;
    w += x * square((*p)->offset - ctx->best->offset);
  }

  ctx->offset = z / y;
  ctx->jitter = sqrt(w / y);
}


/* --- add_peer ------------------------------------------------------------ */

static struct peer *
add_peer(gps_ntp_context_s *ctx, const struct sockaddr *addr, socklen_t len)
{
  struct peer *p;
  struct filter *f;

  p = malloc(sizeof(*p));
  if (!p) return NULL;

  memcpy(p, &(struct peer){
        .next = ctx->peers, .prev = NULL,

        .addrlen = len,

        .shift = 0,

        .stratum = NTP_MAXSTRAT,
        .precision = NTP_MAXDISP,
        .rootdelay = NTP_MAXDISP,
        .rootdisp = NTP_MAXDISP,
        .delay = NTP_MAXDISP,
        .disp = NTP_MAXDISP,
        .jitter = ctx->precision,

        .mindist = { .peer = p, .type = -1 },
        .maxdist = { .peer = p, .type = +1 },

        .burst = NTP_BURST,
        .poll = NTP_BPOLL,
        .nextpoll = getsystime()
        }, sizeof(struct peer));
  memcpy(p->addr, addr, len),
  getnameinfo(addr, len, p->hname, sizeof(p->hname), NULL, 0, 0);
  if (!*p->hname) snprintf(p->hname, sizeof(p->hname), "<unnamed>");
  p->ts = tstofp64(p->nextpoll);
  forarray(f, p->filter)
    *f = (struct filter){
      .offset = 0., .delay = NTP_MAXDISP, .disp = NTP_MAXDISP, .ts = p->ts
    };

  if (ctx->peers) ctx->peers->prev = p;
  ctx->peers = p;
  ctx->npeers++;

  warnx("adding ntp server %s", p->hname);
  return p;
}


/* --- del_peer ------------------------------------------------------------ */

static void
del_peer(gps_ntp_context_s *ctx, struct peer *peer, const char *msg)
{
  warnx("removing ntp server %s: %s", peer->hname, msg);
  peer->nextpoll = (struct timespec){0};
  ctx->npeers--;
}


/* --- find_peer ----------------------------------------------------------- */

static struct peer *
find_peer(gps_ntp_context_s *ctx, const struct sockaddr *addr)
{
  struct peer *p;

  forlist(p, ctx->peers, next) {
    if (p->addr->sa_family != addr->sa_family) continue;
    switch(p->addr->sa_family) {
      default: continue;

      case AF_INET:
        if (((struct sockaddr_in *)p->addr)->sin_addr.s_addr !=
            ((struct sockaddr_in *)addr)->sin_addr.s_addr)
          continue;
        if (((struct sockaddr_in *)p->addr)->sin_port !=
            ((struct sockaddr_in *)addr)->sin_port)
          continue;
        break;

      case AF_INET6:
        if (memcmp(((struct sockaddr_in6 *)p->addr)->sin6_addr.s6_addr,
                   ((struct sockaddr_in6 *)addr)->sin6_addr.s6_addr,
                   16 * sizeof(uint8_t)))
          continue;
        if (((struct sockaddr_in6 *)p->addr)->sin6_port !=
            ((struct sockaddr_in6 *)addr)->sin6_port)
          continue;
        break;
    }

    return p;
  }

  return NULL;
}


/* --- poll_update --------------------------------------------------------- */

static void
poll_update(struct peer *peer)
{
  if (cmpts(peer->nextpoll, ==, (struct timespec){0}))
    return;

  if (!peer->burst && peer->poll < NTP_MINPOLL) peer->poll = NTP_MINPOLL;
  if (peer->poll > NTP_MAXPOLL) peer->poll = NTP_MAXPOLL;

  peer->nextpoll = fp64tots(peer->xmt);
  peer->nextpoll.tv_sec += 1 << peer->poll;
  peer->nextpoll.tv_nsec += random() % 1000000000;
}


/* --- lookup_servers ------------------------------------------------------ */

static int
lookup_servers(gps_ntp_context_s *ctx)
{
  static const int *af, afs[] = { AF_INET6, AF_INET };
  struct addrinfo *res, *res0;
  char **h;
  int s;

  ctx->nextdns.tv_sec += CTX_DNSPOLL;

  forptr(h, ctx->servers, ctx->nservers) {
    /* avoid useless queries if there are too many peers */
    if (ctx->npeers >= CTX_MAXPEERS) return 0;

    forarray(af, afs) {	/* check AF_INET6 first, then AF_INET */
      s = getaddrinfo(
        *h, "ntp", &(struct addrinfo){
          .ai_family = *af,
          .ai_socktype = SOCK_DGRAM,
          .ai_protocol = IPPROTO_UDP }, &res0);
      if (s) continue;

      forlist(res, res0, ai_next) {
        if (ctx->npeers >= CTX_MAXPEERS) break;
        if (find_peer(ctx, res->ai_addr)) continue;

        if (!add_peer(ctx, res->ai_addr, res->ai_addrlen)) return -1;
      }
      freeaddrinfo(res0);
      break;
    }
    if (s) {
      warnx("%s: %s", *h, gai_strerror(s));
      errno = EINVAL;
      return s;
    }
  }

  return 0;
}
